"""Includes functionality for inline editing of SOI."""
from PySide2.QtCore import Qt, QRectF, QTimer, QPoint, QMarginsF
from PySide2.QtWidgets import (
    QApplication,
    QScrollArea,
    QLabel,
    QGraphicsScene,
    QGraphicsView,
    QGraphicsRectItem,
)
from PySide2.QtGui import QFont, QPixmap, QBrush, QPalette, QPainter
from PySide2.QtPrintSupport import QPrinter
from soitool.soi import ModuleLargerThanBinError
from soitool.dialog_wrappers import exec_warning_dialog


class InlineEditableSOIView(QScrollArea):
    """Widget som kan byttes ut med view, edit etc."""

    def is_widget_in_scene(self, widget):
        """Indicate wether given widget already has a proxy in the scene."""
        for proxy in self.proxies:
            if proxy.widget() == widget:
                return True
        return False

    def ensure_proxies(self):
        """Make sure all modules of the SOI have a proxy inside the scene."""
        for module in self.soi.modules:
            if not self.is_widget_in_scene(module["widget"]):
                proxy = self.scene.addWidget(module["widget"])
                self.proxies.add(proxy)

    def mousePressEvent(self, _):
        """Reorganize modules when pressed.

        This is a temporary way to activate reorganization of widgets. Note
        that will not be triggered by clicks on modules in the scene.

        If reorganization cannot occur because a module is too large the user
        will be informed. It's worth noting that with the current
        implementation this is the only way the user will be informed of this,
        so reorganization that is not triggered here will not give feedback to
        the user.
        """
        try:
            self.soi.reorganize()
        except ModuleLargerThanBinError:
            exec_warning_dialog(
                text="Minst én av modulene er for store "
                "for å passe på én side!",
                informative_text="Programmet kan desverre ikke fikse dette "
                "for deg. Se igjennom modulene og sjekk at "
                "alle moduler er mindre enn én enkelt side.",
            )

    def __init__(self, soi):
        super().__init__()

        self.soi = soi

        self.number_of_pages = 1
        self.proxies = set()

        # necessary to make the scroll area fill the space it's given
        self.setWidgetResizable(True)

        # scene and view widgets
        self.scene = QGraphicsScene()
        self.setup_scene()
        self.view = QGraphicsView(self.scene)

        self.setWidget(self.view)

        self.ensure_proxies()
        self.update_pages()

        # add listeners to react properly to SOI changes
        self.soi.add_reorganization_listener(self.update_pages)
        self.soi.add_new_module_listener(self.ensure_proxies)
        self.soi.add_update_property_listener(self.update_pages)

        # self.launch_auto_zoom()

    def produce_pdf(self, filename):
        """Produce PDF using QGraphicsScene.

        Renders the QGraphicsScene-representation of the SOI as a PDF. This
        PDF is in theory supposed to be searchable [1], but in practice it
        seems that each QGraphicsItem in the scene is simply dumped as an
        image.

        Sources:
        [1]: https://doc.qt.io/qt-5/qprinter.html#OutputFormat-enum

        Parameters
        ----------
        filename : str
            Name of file to store PDF in. If file exists it will be
            overwritten. Note that the filename should contain the extension
            '.pdf' to be properly handled by operating systems.
        """
        printer = QPrinter(QPrinter.HighResolution)
        printer.setOutputFormat(QPrinter.PdfFormat)
        printer.setOutputFileName(filename)
        printer.setPageSize(QPrinter.A4)
        printer.setOrientation(QPrinter.Landscape)
        printer.setPageMargins(QMarginsF(0, 0, 0, 0))

        painter = QPainter()

        try:
            ok = painter.begin(printer)

            if not ok:
                raise ValueError(
                    "Not able to begin QPainter using QPrinter "
                    "based on argument "
                    "filename '{}'".format(filename)
                )

            # render each page to own PDF page
            for i in range(self.number_of_pages):

                x = 0
                y = self.soi.HEIGHT * i + self.soi.PADDING * i

                self.scene.render(
                    painter,
                    source=QRectF(x, y, self.soi.WIDTH, self.soi.HEIGHT),
                )

                # if there are more pages, newPage
                if i + 1 < self.number_of_pages:
                    printer.newPage()
        finally:
            painter.end()

    def update_number_of_pages(self):
        """Make sure number of pages excactly fits SOI modules.

        The minimum page count is 1.
        """
        required_pages = 1
        for module in self.soi.modules:
            if module["meta"]["page"] > required_pages:
                required_pages += 1
        self.number_of_pages = required_pages

    def draw_pages(self):
        """Draw self.number_of_pages pages."""
        for i in range(self.number_of_pages):

            x = 0
            y = self.soi.HEIGHT * i + self.soi.PADDING * i

            # adjust page size
            full_scene_height = y + self.soi.HEIGHT
            self.scene.setSceneRect(
                QRectF(0, 0, self.soi.WIDTH, full_scene_height)
            )

            self.draw_page(x, y)
            self.draw_header(x + self.soi.PADDING, y + self.soi.PADDING, i + 1)

        for proxy in self.proxies:
            # redraw of pages requires modules to be moved to front again
            proxy.setZValue(1)

    def update_pages(self):
        """Update pages drawn in the scene to reflect the SOI."""
        self.update_number_of_pages()
        self.draw_pages()

    def draw_header(self, x, y, page_number):
        """Draw header staring at given position.

        Parameters
        ----------
        x : int
        y : int
        """
        # title
        label_title = QLabel(self.soi.title)
        label_title.move(x, y)
        label_title.setStyleSheet("background-color: rgba(0,0,0,0%)")
        label_title.setFont(QFont("Times New Roman", 50))
        self.scene.addWidget(label_title)

        # page numbering
        page_number = QLabel(
            "{} av {}".format(page_number, self.number_of_pages)
        )
        page_number.setStyleSheet("background-color: rgba(0,0,0,0%)")
        page_number.setFont(QFont("Times New Roman", 50))
        # source: https://stackoverflow.com/a/8638114/3545896
        # CAUTION: does not work if font is set through stylesheet
        label_width = (
            page_number.fontMetrics().boundingRect(page_number.text()).width()
        )
        page_number.move(x + (self.soi.CONTENT_WIDTH - label_width) / 2, y)
        self.scene.addWidget(page_number)

        # classification
        classification = QLabel(self.soi.classification)
        classification.setStyleSheet(
            "background-color: rgba(0,0,0,0%); " "color: red"
        )
        classification.setFont(QFont("Times New Roman", 50))
        # source: https://stackoverflow.com/a/8638114/3545896
        # CAUTION: does not work if font is set through stylesheet
        label_width = (
            classification.fontMetrics()
            .boundingRect(classification.text())
            .width()
        )
        x_pos = (
            x + self.soi.CONTENT_WIDTH - label_width - self.soi.HEADER_HEIGHT
        )
        classification.move(x_pos, y)
        self.scene.addWidget(classification)

        # patch
        pixmap = QPixmap(self.soi.icon)
        patch = QLabel()
        patch.setPixmap(
            pixmap.scaled(self.soi.HEADER_HEIGHT, self.soi.HEADER_HEIGHT)
        )
        patch.move(x + self.soi.CONTENT_WIDTH - self.soi.HEADER_HEIGHT, y)
        self.scene.addWidget(patch)

    def draw_page(self, x, y):
        """Draw page starting at given position.

        Parameters
        ----------
        x : int
        y : int
        """
        # color the page white
        page_background = QGraphicsRectItem(
            x, y, self.soi.WIDTH, self.soi.HEIGHT
        )
        page_background.setBrush(QBrush(Qt.white))
        self.scene.addItem(page_background)

        # draw borders
        self.scene.addRect(x, y, self.soi.WIDTH, self.soi.HEIGHT)
        self.scene.addRect(
            x + self.soi.PADDING,
            y + self.soi.PADDING,
            self.soi.CONTENT_WIDTH,
            self.soi.CONTENT_HEIGHT,
        )
        self.scene.addRect(
            x + self.soi.PADDING,
            y + self.soi.PADDING,
            self.soi.CONTENT_WIDTH,
            self.soi.CONTENT_HEIGHT,
        )
        self.scene.addRect(
            x + self.soi.PADDING,
            y + self.soi.PADDING,
            self.soi.CONTENT_WIDTH,
            self.soi.HEADER_HEIGHT,
        )

    def setup_scene(self):
        """Prepare scene for use.

        Draws borders, background, etc.
        """
        # sets the background color of the scene to be the appropriate
        # system-specific background color
        # source: https://stackoverflow.com/a/23880531/3545896
        # source: https://stackoverflow.com/questions/15519749/how-to-get-widget-background-qcolor
        app = QApplication.instance()
        self.scene.setBackgroundBrush(app.palette().color(QPalette.Window))

        self.update_pages()

    def launch_auto_zoom(self):
        """Zoom in a regular interval.

        Used to demonstrate zooming only, should be removed once the project
        matures.
        """

        def do_on_timeout():
            self.zoom(1 / 1.00005)

        timer = QTimer(self)
        timer.timeout.connect(do_on_timeout)
        timer.start(10)

    def zoom(self, zoom_factor):
        """Zoom GraphicsView by zoom_factor.

        source: https://stackoverflow.com/a/29026916/3545896

        Parameters
        ----------
        zoom_factor : float
            > 1 to zoom in and < 1 to zoom out
        """
        # Set Anchors
        self.view.setTransformationAnchor(QGraphicsView.NoAnchor)
        self.view.setResizeAnchor(QGraphicsView.NoAnchor)

        # Zoom
        self.view.scale(zoom_factor, zoom_factor)

        pos_old = self.view.mapToScene(
            QPoint(self.soi.WIDTH / 2, self.soi.HEIGHT / 2)
        )
        pos_new = self.view.mapToScene(self.soi.WIDTH / 2, self.soi.HEIGHT / 2)
        delta = pos_new - pos_old
        self.view.translate(delta.x(), delta.y())
