"""Inclues the main window of our application.

Built up by widgets implemented in other modules.
"""
import sys
import os
from enum import Enum
from functools import partial
from PySide2.QtWidgets import (
    QTabWidget,
    QMainWindow,
    QApplication,
    QAction,
    QFileDialog,
)
from PySide2.QtGui import QIcon
from PySide2.QtCore import QTimer
from soitool.soi_workspace_widget import SOIWorkspaceWidget
from soitool.codebook_to_pdf import generate_codebook_pdf
from soitool.dialog_wrappers import exec_info_dialog
from soitool.codebook_widget import CodebookWidget
from soitool.codebook_model_view import CodeBookTableModel
from soitool.database import Database, DBPATH
from soitool.help_actions import ShortcutsHelpDialog, BasicUsageHelpDialog
from soitool.serialize_export_import_soi import (
    export_soi,
    import_soi,
)


class ModuleType(Enum):
    """Enumerate with types of modules."""

    MAIN_MODULE = 0
    ATTACHMENT_MODULE = 1


class MainWindow(QMainWindow):
    """MainWindow, shell of the entire application.

    Parameter 'db_path' is the path to an existing or future database-file.
    """

    def __init__(self, db_path=DBPATH):  # pylint: disable = R0914, R0915
        super().__init__()
        self.setGeometry(100, 100, 800, 800)
        self.setWindowTitle("SOI-tool")
        self.statusBar()

        self.popup_shortcut_help = ShortcutsHelpDialog()
        self.popup_basic_use_help = BasicUsageHelpDialog()
        # Database instance
        self.database = Database(db_path)
        # Timer for automatic update of codes in CodeBook
        self.timer = QTimer()
        # Interval i set to msec since last 24h update
        self.timer.setInterval(
            self.database.seconds_to_next_update(60 * 60 * 24) * 1000
        )
        self.timer.timeout.connect(partial(self.regenerate_codes, auto=True))
        self.timer.start()

        # Add SOIWorkspaceWidget-tab
        self.tabs = QTabWidget()
        self.tabs.setTabsClosable(True)

        tab = SOIWorkspaceWidget()
        self.tabs.addTab(tab, "MainTab")
        self.setCentralWidget(self.tabs)

        self.setup_menubar()

        # Add HV logo
        filename = "media/HVlogo.PNG"
        dirname = os.path.dirname(__file__)
        filepath = os.path.join(dirname, filename)
        self.setWindowIcon(QIcon(filepath))

        self.action_shortcut_help.triggered.connect(self.open_shortcut_help)
        self.action_basic_use_help.triggered.connect(self.open_basic_use_help)

    # pylint: disable=R0914, R0915
    # Ignoring "Too many local variables" and "Too many statements"
    def setup_menubar(self):
        """Set up menubar with submenus and actions."""
        menu = self.menuBar()
        file_menu = menu.addMenu("SOI")
        codebook_menu = menu.addMenu("Kodebok")
        help_menu = menu.addMenu("Hjelp")

        # New SOI
        new_soi = QAction("Ny", self)
        new_soi.setShortcut("Ctrl+n")
        new_soi.setStatusTip("Opprett en ny SOI")
        file_menu.addAction(new_soi)
        # new_soi.triggered.connect(self.<nySOI>)

        # Open file
        open_file = QAction("Åpne", self)
        open_file.setShortcut("Ctrl+o")
        open_file.setStatusTip("Åpne en SOI fra fil")
        open_file.triggered.connect(self.import_soi)
        file_menu.addAction(open_file)

        # Open file from DB
        open_file_db = QAction("Åpne fra DB", self)
        open_file_db.setShortcut("Ctrl+d")
        open_file_db.setStatusTip("Åpne en SOI fra databasen")
        file_menu.addAction(open_file_db)

        # Preview SOI
        preview_soi = QAction("Forhåndsvis")
        preview_soi.setShortcut("Ctrl+p")
        preview_soi.setStatusTip("Forhåndsvis SOI som PDF")
        file_menu.addAction(preview_soi)

        # Save to DB
        save_soi = QAction("Lagre i DB", self)
        save_soi.setShortcut("Ctrl+s")
        save_soi.setStatusTip("Lagre SOI i databasen")
        file_menu.addAction(save_soi)

        # Export SOI
        export_serialized_soi = file_menu.addMenu("Eksporter")
        # Compressed SOI
        export_compressed = QAction("Komprimert", self)
        export_compressed.setShortcut("Ctrl+e")
        export_compressed.setStatusTip("Eksporter komprimert SOI")
        export_compressed.triggered.connect(
            partial(self.try_export_soi, compressed=True)
        )
        # Uncompressed SOI
        export_uncompressed = QAction("Ukomprimert", self)
        export_uncompressed.setStatusTip("Eksporter ukomprimert SOI")
        export_uncompressed.triggered.connect(
            partial(self.try_export_soi, compressed=False)
        )
        export_serialized_soi.addAction(export_compressed)
        export_serialized_soi.addAction(export_uncompressed)
        file_menu.addMenu(export_serialized_soi)

        # View/edit Codebook
        codebook = QAction("Se/rediger kodebok", self)
        codebook.setStatusTip("Se/rediger kodebok")
        codebook.triggered.connect(self.open_codebook_tab)
        codebook_menu.addAction(codebook)

        # Regenerate codebook-codes:
        regenerate_codes = QAction("Nye koder i db", self)
        regenerate_codes.setStatusTip("Nye koder lages tilfeldig")
        regenerate_codes.triggered.connect(self.regenerate_codes)
        codebook_menu.addAction(regenerate_codes)

        # Export codebook as PDF
        export_codebook = codebook_menu.addMenu("Eksporter")

        # Export full codebook
        export_codebook_full = QAction("Stor kodebok", self)
        export_codebook_full.setStatusTip("Eksporter stor kodebok som PDF")
        export_codebook_full.triggered.connect(
            partial(generate_codebook_pdf, database=self.database)
        )
        export_codebook.addAction(export_codebook_full)

        # Export small codebook
        export_codebook_small = QAction("Liten kodebok", self)
        export_codebook_small.setStatusTip("Eksporter liten kodebok som PDF")
        export_codebook_small.triggered.connect(
            partial(generate_codebook_pdf, database=self.database, small=True)
        )
        export_codebook.addAction(export_codebook_small)

        # Hot keys
        self.action_shortcut_help = QAction("Hurtigtaster", self)
        self.action_shortcut_help.setStatusTip(
            "Vis oversikt over hurtigtaster"
        )
        help_menu.addAction(self.action_shortcut_help)

        # Easy use
        self.action_basic_use_help = QAction("Enkel bruk", self)
        self.action_basic_use_help.setStatusTip(
            "Vis enkel bruk av programvaren"
        )
        help_menu.addAction(self.action_basic_use_help)

    def open_codebook_tab(self):
        """Open tab containing CodeBookTableView and CodebookRowAdder.

        Select codebook-tab if it is already open,
        create and select codebook-tab if not open.
        """
        # Loop through tabs to look for existing codebook-tab:
        for i in range(self.tabs.count()):
            if self.tabs.tabText(i) == "Kodebok":
                self.tabs.setCurrentIndex(i)
                break
        # Codebook-tab does not exist, create, add and select tab
        else:
            tab = CodebookWidget(self.database)
            self.tabs.addTab(tab, "Kodebok")
            self.tabs.setCurrentWidget(tab)

    def try_export_soi(self, compressed=True):
        """Export the SOI in the current tab.

        Feedback is given through a dialog if the current tab does not contain
        an SOI (tab is not a SOIWorkspaceWidget).

        Parameters
        ----------
        compressed : bool, optional
            Serialized SOI is compressed if True (default)
        """
        tab_widget = self.tabs.currentWidget()

        # If tab contains an SOI
        if isinstance(tab_widget, SOIWorkspaceWidget):
            export_soi(tab_widget.soi, compressed)
        else:
            exec_info_dialog(
                "Valgt tab er ingen SOI-tab",
                "Den valgte taben inneholder ingen SOI.\n"
                "For å eksportere en SOI må riktig tab velges.",
            )

    def import_soi(self):
        """Import serialized SOI.

        Launches a QFileDialog with a name-filter, where .txt and .json are
        accepted file extensions.
        A SOIWorkspaceWidget containing the SOI-object is created and opened
        in a new tab, which is selected.
        """
        # Get file-path from dialog
        file_path = QFileDialog().getOpenFileName(
            self,
            "Åpne SOI",
            os.getcwd(),
            "Text/JSON-filer (SOI_*.txt SOI_*.json)",
        )[0]

        if len(file_path) > 0:
            soi = import_soi(file_path)

            # Create and select tab
            tab = SOIWorkspaceWidget(soi)
            self.tabs.addTab(tab, soi.title)
            self.tabs.setCurrentWidget(tab)

    def regenerate_codes(self, auto=False):
        """Regenerate codebook-codes and update codebook-tab if open.

        Parameters
        ----------
        auto : bool, optional
            True if self.timer called this function, by default False.
        """
        if auto:

            def regenerate():
                self.database.update_codebook_auto(self.timer)

        else:

            def regenerate():
                self.database.update_codebook()

        for i in range(self.tabs.count()):
            if self.tabs.tabText(i) == "Kodebok":
                view = self.tabs.widget(i).view
                view.setModel(None)
                regenerate()
                view.setModel(CodeBookTableModel())
                break
        else:
            regenerate()

    def open_shortcut_help(self):
        """Open shortcut dialog."""
        self.popup_shortcut_help.setWindowTitle("Hurtigtaster")
        self.popup_shortcut_help.exec()

    def open_basic_use_help(self):
        """Open basic usage dialog."""
        self.popup_basic_use_help.setWindowTitle("Enkel bruk")
        self.popup_basic_use_help.exec()


if __name__ == "__main__":

    app = QApplication(sys.argv)
    WINDOW = MainWindow()
    WINDOW.showMaximized()
    app.exec_()
