"""Module containing parent-class for code-table-modules.

Parent-class for AuthenticationBoardModule and SubtractorCodesModule.
"""
from PySide2.QtWidgets import QTableWidget, QTableWidgetItem
from PySide2.QtCore import Qt
from soitool.coder import get_code_set, get_code
from soitool.modules.module_base import (
    ModuleBase,
    get_table_size,
    resize_table,
    HEADLINE_FONT,
)

AUTHENTICATIONBOARDMODULE = "AuthenticationBoardModule"
SUBTRACTORCODESMODULE = "SubtractorcodesModule"


class Meta(type(ModuleBase), type(QTableWidget)):
    """Used as a metaclass to enable multiple inheritance."""


class CodeTableBase(ModuleBase, QTableWidget, metaclass=Meta):
    """Parent-class for AuthenticationBoardModule and SubtractorcodesModule.

    Inherits from ModuleBase and QTableWidget.
    ModuleBase is used as an interface, it's methods are overridden.
    """

    def __init__(self, size, data):
        if self.type not in [
            AUTHENTICATIONBOARDMODULE,
            SUBTRACTORCODESMODULE,
        ]:
            raise ValueError(
                "Invalid value for class-variable type: "
                "'{}'".format(self.type)
            )
        QTableWidget.__init__(self)
        ModuleBase.__init__(self)

        # Remove headers and scrollbars
        self.horizontalHeader().hide()
        self.verticalHeader().hide()
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

        # Resize table when headline changes
        self.cellChanged.connect(
            lambda: resize_table(
                self,
                resize_rows=False,
                resize_columns=False,
                has_headline=True,
            )
        )
        # If parameters are None, generate new table
        if size is None and data is None:
            self.generate_table()
            self.resizeColumnsToContents()
            self.insert_headline(self.headline)

            # Resize height of rows and set size of window
            resize_table(self, resize_columns=False, has_headline=True)
        else:
            self.setColumnCount(len(data[1]))
            self.setRowCount(len(data) - 1)  # - 1 to skip headline

            # Set cell-items
            for i in range(self.rowCount()):
                for j in range(self.columnCount()):
                    item = QTableWidgetItem(data[i + 1][j])  # +1 skip headline
                    item.setTextAlignment(Qt.AlignCenter)
                    self.setItem(i, j, item)

            self.resizeColumnsToContents()
            self.insert_headline(data[0])

            resize_table(self, resize_columns=False, has_headline=True)
            self.setFixedWidth(size["width"])
            self.setFixedHeight(size["height"])

    def generate_table(self):
        """Insert row identifiers and codes."""
        # Set number of rows and columns
        self.setRowCount(self.start_no_of_codes)
        self.setColumnCount(3)
        self.insert_row_identifiers()

        # Generate codes
        codes = list(
            get_code_set(
                self.start_no_of_codes,
                self.code_length,
                self.code_character_type,
                self.space_interval,
                self.space_amount,
            )
        )
        # Insert codes
        for i in range(self.start_no_of_codes):
            # Insert non-editable code in third column
            item_third = QTableWidgetItem(codes[i])
            if self.type == "SubtractorcodesModule":
                item_third.setTextAlignment(Qt.AlignCenter)
            item_third.setFlags(item_third.flags() ^ Qt.ItemIsEditable)
            self.setItem(i, 2, item_third)

    def insert_headline(self, text=None):
        """Insert headline text.

        Parameters
        ----------
        text : string, optional
            The headline text, self.headline is used if None,
            by default None.
        """
        headline = self.headline if text is None else text

        item_headline = QTableWidgetItem(headline)
        item_headline.setTextAlignment(Qt.AlignHCenter)
        item_headline.setFont(HEADLINE_FONT)

        self.insertRow(0)
        self.setItem(0, 0, item_headline)
        self.setSpan(0, 0, 1, self.columnCount())  # Make cell span all columns

    def generate_unique_authentication_code(self):
        """Generate authentication-code that does not already exist.

        Returns
        -------
        string
            Generated, unique authentication-code.
        """
        # Get existing codes
        existing_codes = self.get_codes()

        # Randomly generate a new code until it is unique
        unique_code = False
        while not unique_code:
            code = get_code(
                self.code_length,
                self.code_character_type,
                self.space_interval,
                self.space_amount,
            )
            unique_code = code not in existing_codes

        return code

    def get_codes(self):
        """Get all codes in table.

        Returns
        -------
        List
            List containing codes.
        """
        codes = []

        # Start from 1 to skip headline-row
        for i in range(1, self.rowCount()):
            codes.append(self.item(i, 2).text())

        return codes

    def keyPressEvent(self, event):
        """Add or remove row when 'Ctrl + +' and 'Ctrl + -' are pressed."""
        if (
            event.modifiers() == Qt.ControlModifier
            and event.key() == Qt.Key_Plus
        ):
            # If a row is selected
            row_index = self.currentRow()
            if row_index != -1:
                self.add_row(row_index)
        elif (
            event.modifiers() == Qt.ControlModifier
            and event.key() == Qt.Key_Underscore
        ):
            # If at least one row (+ headline-row) exists and a row other than
            # headline-row is selected
            row_index = self.currentRow()
            if self.rowCount() > 2 and row_index != 0 and row_index != -1:
                self.remove_row(row_index)
        else:
            super().keyPressEvent(event)

    def get_size(self):
        """Return size of table."""
        return get_table_size(self)

    def get_data(self):
        """Return list containing all data.

        Returns
        -------
        List
            List[0] contains headline,
            list[x][y] represents value of row x, column y.
        """
        data = []
        item_headline = self.item(0, 0)
        if item_headline is not None:
            data.append(item_headline.text())

        for i in range(1, self.rowCount()):
            row = []
            for j in range(self.columnCount()):
                row.append(self.item(i, j).text())
            data.append(row)

        return data

    @staticmethod
    def get_icon():
        """Abstract method, should be implemented by derived class."""
        raise NotImplementedError

    @staticmethod
    def get_user_friendly_name():
        """Abstract method, should be implemented by derived class."""
        raise NotImplementedError
