"""Base/interface of each module."""
from abc import ABC
from PySide2 import QtGui

# Font for module headline
HEADLINE_FONT = QtGui.QFont()
HEADLINE_FONT.setFamily("Arial")
HEADLINE_FONT.setPointSize(12)
HEADLINE_FONT.setWeight(100)


class ModuleBase(ABC):
    """Interface for SOI-modules."""

    type = None

    def __init__(self):
        """Class-variable 'type' should be set by derived class."""
        if self.type is None:
            raise NotImplementedError

    def get_size(self):
        """Abstract method, should be implemented by derived class."""
        raise NotImplementedError

    def get_data(self):
        """Abstract method, should be implemented by derived class."""
        raise NotImplementedError

    @staticmethod
    def get_user_friendly_name():
        """Abstract method, should be implemented by derived class."""
        raise NotImplementedError

    @staticmethod
    def get_icon():
        """Abstract method, should be implemented by derived class."""
        raise NotImplementedError


def resize_table(
    table, resize_rows=True, resize_columns=True, has_headline=False
):
    """Resize a given QTableWidget.

    Parameters
    ----------
    table : QTableWidget
        QTablewidget-instance to resize.
    resize_rows : bool
        Resizes rows to contents if True, by default True.
    resize_columns : bool
        Resizes columns to contents if True, by default True.
    has_headline : bool
        True if the table has a headline, by default False.
        Last column is widened if headline is wider than table.
    """
    if resize_columns:
        table.resizeColumnsToContents()
    if resize_rows:
        table.resizeRowsToContents()

    # If table has a headline, make sure table is wide enough to fit it.
    if has_headline:
        last_column_index = table.columnCount() - 1
        table.resizeColumnToContents(last_column_index)
        width, height = get_table_size(table)

        # Get width of headline
        headline = table.item(0, 0).text()
        headline_width = (
            QtGui.QFontMetricsF(HEADLINE_FONT).horizontalAdvance(headline) + 10
        )
        # If headline is wider than table
        if width < headline_width:
            difference = headline_width - width
            width += difference
            old_width = table.columnWidth(last_column_index)
            table.setColumnWidth(last_column_index, old_width + difference)
    else:
        width, height = get_table_size(table)

    table.setFixedWidth(width)
    table.setFixedHeight(height)


def get_table_size(widget):
    """Calculate and return total width and height of a QTableWidget.

    Parameters
    ----------
    widget : QTableWidget
        QTableWidget-instance to calculate and return size of.

    Returns
    -------
    Tuple
        Total (width, height)
    """
    width = 0
    height = 0

    for i in range(widget.columnCount()):
        width += widget.columnWidth(i) + 0.5

    for i in range(widget.rowCount()):
        height += widget.rowHeight(i) + 0.5

    return width, height


def set_module_pos(widget, pos):
    """Set position of module (widget).

    Parameters
    ----------
    widget : QWidget
        Widget to move.
    pos : QPoint
        Position (x, y).
    """
    widget.move(pos)
