"""SOI module for arbitrary text.

Sources:
* https://stackoverflow.com/questions/47710329/how-to-adjust-qtextedit-to-fit-its-contents
* https://stackoverflow.com/questions/3050537/resizing-qts-qtextedit-to-match-text-height-maximumviewportsize
"""
from PySide2.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QLineEdit,
    QTextEdit,
    QSizePolicy,
)
from PySide2.QtCore import QSize, Qt
from PySide2.QtGui import QIcon, QFontMetricsF
from soitool.modules.module_base import ModuleBase, HEADLINE_FONT


class TextEditWithSizeOfContent(QTextEdit):
    """Custom QTextEdit subclass that grows to fit it's text."""

    def __init__(self, *arg, **kwargs):
        super(TextEditWithSizeOfContent, self).__init__(*arg, **kwargs)

        self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setLineWrapMode(QTextEdit.NoWrap)

    def sizeHint(self):
        """Give size hint using width of text."""
        size = QSize(self.document().size().toSize())
        # NOTE that the following is not respected for dimensions below 100,100
        size.setWidth(max(100, size.width()))
        size.setHeight(max(100, size.height()))
        return size

    def resizeEvent(self, event):
        """Update geometry before handling the resizeEvent.

        See sources in module docstring.

        Parameters
        ----------
        event : QResizeEvent
            event sent by Qt
        """
        self.updateGeometry()
        super(TextEditWithSizeOfContent, self).resizeEvent(event)


class LineEditWithSizeOfContent(QLineEdit):
    """Custom QLineEdit subclass that grows to fit it's text."""

    def __init__(self, *arg, **kwargs):
        super(LineEditWithSizeOfContent, self).__init__(*arg, **kwargs)
        self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)

    def sizeHint(self):
        """Give size hint using width of text."""
        size_hint_parent = super(LineEditWithSizeOfContent, self).sizeHint()
        content_width = QFontMetricsF(self.fontMetrics()).horizontalAdvance(
            self.text()
        )
        # Hardcoded addition because font metrics is slightly off...
        # Likely because of the font weight
        content_width += 10
        return QSize(content_width, size_hint_parent.height())

    def resizeEvent(self, event):
        """Update geometry before handling the resizeEvent.

        See sources in module docstring.

        Parameters
        ----------
        event : QResizeEvent
            event sent by Qt
        """
        self.updateGeometry()
        super(LineEditWithSizeOfContent, self).resizeEvent(event)


class Meta(type(ModuleBase), type(QWidget)):
    """Used as a metaclass to enable multiple inheritance."""


class FreeTextModule(ModuleBase, QWidget, metaclass=Meta):
    """Module for arbitrary text.

    ## Note about widget size

    This widget might be too small to be shown in a window by itself. If you
    need to show only this widget, you should show it by wrapping it in a
    QWidget.

    Parameters
    ----------
    size
        Not used.
    data : list
        First index should be header, second index should be body
    """

    # Ignoring Pylint's "Unused argument 'size'" error because it is
    # intentionally left unused
    # pylint: disable=W0613
    def __init__(self, size=None, data=None):
        self.type = "FreeTextModule"
        QWidget.__init__(self)
        ModuleBase.__init__(self)

        self.line_edit_header = LineEditWithSizeOfContent()
        self.line_edit_header.setFont(HEADLINE_FONT)
        self.text_edit_body = TextEditWithSizeOfContent()

        # When the contents of these widgets change we need to manually trigger
        # adjust of size, even on self. Without adjust of size on self the
        # widget will leave behind gray color when it is shrunk in the
        # QGraphicsScene
        self.line_edit_header.textChanged.connect(
            lambda: (self.line_edit_header.adjustSize(), self.adjustSize())
        )
        self.text_edit_body.textChanged.connect(
            lambda: (self.text_edit_body.adjustSize(), self.adjustSize())
        )

        self.layout = QVBoxLayout()
        self.layout.setAlignment(Qt.AlignLeft | Qt.AlignTop)
        self.layout.setSpacing(0)
        self.layout.setMargin(0)
        self.layout.addWidget(self.line_edit_header)
        self.layout.addWidget(self.text_edit_body)
        self.setLayout(self.layout)

        if data is not None:
            header, body = data[0], data[1]
            self.line_edit_header.setText(header)
            self.text_edit_body.setText(body)

    def resizeEvent(self, event):
        """Update geometry before handling the resizeEvent.

        See sources in module docstring.

        Parameters
        ----------
        event : QResizeEvent
            event sent by Qt
        """
        self.updateGeometry()
        super(FreeTextModule, self).resizeEvent(event)

    def get_size(self):
        """Get size of widget.

        Returns
        -------
        Tuple
            (width, height)
        """
        size = self.sizeHint()
        return (size.width(), size.height())

    def get_data(self):
        """Return list containing module data.

        Returns
        -------
        list of module content
            first index contains header, second index contains body
        """
        return [
            self.line_edit_header.text(),
            self.text_edit_body.toPlainText(),
        ]

    @staticmethod
    def get_user_friendly_name():
        """Get user-friendly name of module."""
        return "Fritekst"

    @staticmethod
    def get_icon():
        """Get icon of module."""
        return QIcon("soitool/media/freetextmodule.png")
