"""Module containing SOI-module 'Subtraktorkoder'."""
import string
from PySide2.QtWidgets import QTableWidgetItem
from PySide2 import QtGui
from PySide2.QtCore import Qt
from soitool.modules.module_base import resize_table
from soitool.modules.code_table_base import CodeTableBase

# Maximum number of codes is len(ROW_IDENTIFIERS)/2 columns = 13
START_NO_OF_CODES = 7
CODE_LENGTH = 8

# Characters for first and second column
ROW_IDENTIFIERS = string.ascii_uppercase

# Adds space between sets of characters, 0 => no spaces.
# If code is 12345678 and interval is 2, code will be 1234 5678
SPACE_INTERVAL = 4
SPACE_AMOUNT = 5

HEADLINE_TEXT = "Subtraktorkoder"


class SubtractorcodesModule(CodeTableBase):
    """Modified QTablewidget representing SOI-module 'Subtraktorkoder'.

    By default, the widget initializes with a headline, a row-count of
    START_NO_OF_CODES and three columns.
    If there are 2 rows and ROW_IDENTIFIERS is the alphabet, the first
    column will contain A and B, and the second column will contain C and D.
    The third column contains subtractorcodes of length CODE_LENGTH, spaced out
    for readability if SPACE_INTERVAL and SPACE_AMOUNT larger than 0.

    If parameters are given, the widget initializes accordingly:
    'size' is a dict: {"width": int, "height": int},
    'data' is a 2D list where data[0] is the headline,
    and data[x][y] represents the value in row x, column y.

    The widget does not use more room than needed, and resizes dynamically.
    It has shortcuts for adding and removing rows.
    """

    def __init__(self, size=None, data=None):
        self.type = "SubtractorcodesModule"

        if START_NO_OF_CODES > 13:
            raise ValueError(
                "Invalid value for CONSTANT 'START_NO_OF_CODES': "
                "'{}'".format(START_NO_OF_CODES)
            )
        self.headline = HEADLINE_TEXT
        self.code_length = CODE_LENGTH
        self.start_no_of_codes = START_NO_OF_CODES
        self.space_interval = SPACE_INTERVAL
        self.space_amount = SPACE_AMOUNT
        self.code_character_type = "digits"

        CodeTableBase.__init__(self, size, data)

    def insert_row_identifiers(self, has_headline=False):
        """Insert row identifiers in first and second column.

        Parameters
        ----------
        has_headline : bool, optional
            True if a headline-row exists, by default False.
        """
        start_row = 1 if has_headline else 0

        # Get row identifiers (start-index to skip headline)
        # for column one and two
        last_id_index_one = self.rowCount() - start_row
        start_id_index_two = self.rowCount() - start_row
        last_id_index_two = self.rowCount() * 2 - 2 * start_row

        # If table has 3 rows and ROW_IDENTIFIERS are A-Z,
        # identifiers_one = A, B, C and identifiers_two = C, D, E
        identifiers_one = ROW_IDENTIFIERS[0:last_id_index_one]
        identifiers_two = ROW_IDENTIFIERS[start_id_index_two:last_id_index_two]

        # Insert identifiers in column one and two
        for i in range(2):
            for j in range(start_row, self.rowCount()):
                text = (
                    identifiers_one[j - start_row]
                    if i == 0
                    else identifiers_two[j - start_row]
                )
                item = QTableWidgetItem(text)
                item.setTextAlignment(Qt.AlignCenter)
                item.setFlags(item.flags() ^ Qt.ItemIsEditable)
                self.setItem(j, i, item)

    def add_row(self, selected_row_index):
        """Insert row below the selected row and add data.

        Parameters
        ----------
        selected_row_index : int
            Index of the selected row.
        """
        # If maximum amount of rows not reached (- 1 to skip headline)
        if self.rowCount() - 1 < len(ROW_IDENTIFIERS) / 2:

            # Generate unique code and insert row
            code = self.generate_unique_authentication_code()
            self.insertRow(selected_row_index + 1)

            # Insert row identifiers
            self.insert_row_identifiers(has_headline=True)

            # Insert code
            item_code = QTableWidgetItem(code)
            item_code.setTextAlignment(Qt.AlignCenter)
            item_code.setFlags(item_code.flags() ^ Qt.ItemIsEditable)
            self.setItem(selected_row_index + 1, 2, item_code)

            resize_table(self, resize_columns=False, has_headline=True)

    def remove_row(self, row_index):
        """Remove the selected row.

        Parameters
        ----------
        row_index : int
            Index of the row to remove.
        """
        self.removeRow(row_index)

        self.insert_row_identifiers(has_headline=True)
        resize_table(self, resize_columns=False, has_headline=True)

    @staticmethod
    def get_user_friendly_name():
        """Get user-friendly name of module."""
        return "Subtraktorkoder"

    @staticmethod
    def get_icon():
        """Get icon of module."""
        return QtGui.QIcon("soitool/media/subtractorcodesmodule.png")
