"""Module containing a general SOI-module table."""
from PySide2.QtWidgets import QTableWidget, QTableWidgetItem
from PySide2 import QtGui, QtCore
from soitool.modules.module_base import (
    ModuleBase,
    resize_table,
    get_table_size,
    HEADLINE_FONT,
)

START_ROWS = 2
START_COLUMNS = 2


class Meta(type(ModuleBase), type(QTableWidget)):
    """Used as a metaclass to enable multiple inheritance."""


class TableModule(ModuleBase, QTableWidget, metaclass=Meta):
    """Modified QTableWidget.

    By default, the widget initializes as an empty START_ROWS * START_COLUMNS
    table. If parameters are given, the table initializes accordingly:
    'size' is a dict: {"width": int, "height": int},
    'data' is a 2D list where data[x][y] represents row x, column y.

    The widget does not use more room than needed, and resizes dynamically.
    Columnheaders are styled with light grey background and bold text.
    It has shortcuts for adding and removing rows and columns.

    Inherits from ModuleBase and QTableWidget.
    ModuleBase is used as an interface, it's methods are overridden.
    """

    def __init__(self, size=None, data=None):
        self.type = "TableModule"
        QTableWidget.__init__(self)
        ModuleBase.__init__(self)

        # Remove headers and scrollbars
        self.horizontalHeader().hide()
        self.verticalHeader().hide()
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)

        # If parameters are None, start as empty table.
        if size is None and data is None:
            # Set number of columns and rows
            self.setColumnCount(START_COLUMNS)
            self.setRowCount(START_ROWS)

            # Resize width and height of rows, columns and window
            resize_table(self, resize_rows=True, resize_columns=True)

            # Set header-items
            for i in range(self.columnCount()):
                self.set_header_item(i, "")
        else:
            self.setColumnCount(len(data[0]))
            self.setRowCount(len(data))

            # Set header-items
            for i in range(self.columnCount()):
                self.set_header_item(i, data[0][i])

            # Set cell-items
            for i in range(1, self.rowCount()):
                for j in range(self.columnCount()):
                    item = QTableWidgetItem(data[i][j])
                    self.setItem(i, j, item)

            self.resizeColumnsToContents()
            self.resizeRowsToContents()
            self.setFixedWidth(size["width"])
            self.setFixedHeight(size["height"])

        self.cellChanged.connect(
            lambda: resize_table(self, resize_rows=True, resize_columns=True)
        )

    def keyPressEvent(self, event):
        """Launch actions when specific combinations of keys are pressed.

        If the keys pressed are not related to a shortcut on this custom widget
        the event is sent on to be handled by the superclass (for navigation
        with arrow-keys for.eg.)

        Parameters
        ----------
        event : QKeyEvent
            event sent by Qt for us to handle
        """
        if event.key() == QtCore.Qt.Key_Question:
            self.add_column()
        elif (
            event.modifiers() == QtCore.Qt.ShiftModifier
            and event.key() == QtCore.Qt.Key_Underscore
        ):
            self.remove_column()
        elif (
            event.modifiers() == QtCore.Qt.ControlModifier
            and event.key() == QtCore.Qt.Key_Plus
        ):
            self.add_row()
        elif (
            event.modifiers() == QtCore.Qt.ControlModifier
            and event.key() == QtCore.Qt.Key_Underscore
        ):
            self.remove_row()
        else:
            super(TableModule, self).keyPressEvent(event)

    def set_header_item(self, column, text):
        """Insert item with header-style.

        Item will always be set on header (top) row.

        Parameters
        ----------
        column : int
            What column index for inserting item.
        text : String
            What text the item should contain.
        """
        item = QTableWidgetItem(text)
        item.setBackground(QtGui.QBrush(QtGui.QColor(220, 220, 220)))
        item.setFont(HEADLINE_FONT)
        self.setItem(0, column, item)

    def add_column(self):
        """Add column to the right of selected column."""
        self.insertColumn(self.currentColumn() + 1)
        self.set_header_item(self.currentColumn() + 1, "")
        resize_table(self, resize_rows=True, resize_columns=True)

    def remove_column(self):
        """Remove selected column if two or more columns exist."""
        if self.columnCount() > 1:
            self.removeColumn(self.currentColumn())
            resize_table(self, resize_rows=True, resize_columns=True)

    def add_row(self):
        """Add row below selected row."""
        self.insertRow(self.currentRow() + 1)
        resize_table(self, resize_rows=True, resize_columns=True)

    def remove_row(self):
        """Remove selected row if two or more rows exist (including header)."""
        if self.rowCount() > 2 and self.currentRow() != 0:
            self.removeRow(self.currentRow())
            resize_table(self, resize_rows=True, resize_columns=True)

    def get_size(self):
        """Return size of widget."""
        return get_table_size(self)

    def get_data(self):
        """Return list containing module data.

        Returns
        -------
        List (2D)
            list[x][y] represents value of row x, column y.
        """
        content = []
        for i in range(self.rowCount()):
            row = []
            for j in range(self.columnCount()):
                item = self.item(i, j)
                if item is not None:
                    row.append(item.text())
                else:
                    row.append("")
            content.append(row)

        return content

    @staticmethod
    def get_user_friendly_name():
        """Get user-friendly name of module."""
        return "Generisk tabell"

    @staticmethod
    def get_icon():
        """Get icon of module."""
        return QtGui.QIcon("soitool/media/tablemodule.png")
