"""Includes functionality for serializing, exporting and importing SOI."""
import json
from datetime import datetime
from schema import Schema, And, Or
from soitool.soi import SOI
from soitool.compressor import compress, decompress
from soitool.modules.module_table import TableModule
from soitool.modules.module_authentication_board import (
    AuthenticationBoardModule,
)
from soitool.modules.module_subtractorcodes import SubtractorcodesModule
from soitool.modules.module_freetext import FreeTextModule

# Valid schema for serialized SOI
SERIALIZED_SOI_SCHEMA = Schema(
    {
        "title": And(str, len),
        "description": str,
        "version": And(str, len),
        "date": Or(str, None),
        "valid": {"from_date": Or(str, None), "to_date": Or(str, None)},
        "icon": Or(str, None),
        "classification": And(
            str,
            len,
            Or(
                "UGRADERT",
                "BEGRENSET",
                "KONFIDENSIELT",
                "HEMMELIG",
                "STRENGT HEMMELIG",
            ),
        ),
        "orientation": And(str, len, Or("portrait", "landscape")),
        "placement_strategy": And(str, len, Or("manual", "auto")),
        "algorithm_bin": And(str, len, Or("BFF", "BBF")),
        "algorithm_pack": And(
            str, len, Or("MaxRectsBl", "SkylineBl", "GuillotineBssfSas")
        ),
        "algorithm_sort": And(str, len, Or("none", "area", "width", "height")),
        "modules": [
            {
                "type": And(str, len),
                "data": object,
                "size": {
                    "width": And(Or(int, float), lambda w: w > 0),
                    "height": And(Or(int, float), lambda h: h > 0),
                },
                "meta": {
                    "x": And(Or(int, float), lambda x: x >= 0),
                    "y": And(Or(int, float), lambda y: y >= 0),
                    "page": And(int, lambda page: page >= 0),
                    "name": And(str, len),
                },
            }
        ],
        "attachments": [
            {
                "type": And(str, len),
                "data": object,
                "size": {
                    "width": And(Or(int, float), lambda w: w > 0),
                    "height": And(Or(int, float), lambda h: h > 0),
                },
                "meta": {
                    "x": And(Or(int, float), lambda x: x >= 0),
                    "y": And(Or(int, float), lambda y: y >= 0),
                    "page": And(int, lambda page: page >= 0),
                    "name": And(str, len),
                },
            }
        ],
    }
)


def serialize_soi(soi):
    """Serialize SOI to JSON-string.

    Parameters
    ----------
    soi : soitool.soi.SOI
        SOI to serialize.

    Returns
    -------
    String
        JSON-string containing all SOI-information.

    Raises
    ------
    TypeError
        Raises error if parameter 'soi' is not an SOI.
    """
    # If parameter 'soi' is not an SOI
    if not isinstance(soi, SOI):
        raise TypeError(
            "Invalid type for parameter 'soi': " + "'{}'.".format(soi)
        )

    # Create dict with relevant module-information
    modules = []
    for module in soi.modules:
        width, height = module["widget"].get_size()
        modules.append(
            {
                "type": module["widget"].type,
                "data": module["widget"].get_data(),
                "size": {"width": width, "height": height},
                "meta": module["meta"],
            }
        )

    # Create dict with relevant attachment-information
    attachments = []
    for attachment in soi.attachments:
        width, height = attachment["widget"].get_size()
        attachments.append(
            {
                "type": attachment["widget"].type,
                "data": attachment["widget"].get_data(),
                "size": {"width": width, "height": height},
                "meta": attachment["meta"],
            }
        )

    # Create dict with all relevant SOI-information
    serialized = {
        "title": soi.title,
        "description": soi.description,
        "version": soi.version,
        "date": soi.date,
        "valid": {"from_date": soi.valid_from, "to_date": soi.valid_to},
        "icon": soi.icon,
        "classification": soi.classification,
        "orientation": soi.orientation,
        "placement_strategy": soi.placement_strategy,
        "algorithm_bin": soi.algorithm_bin,
        "algorithm_pack": soi.algorithm_pack,
        "algorithm_sort": soi.algorithm_sort,
        "modules": modules,
        "attachments": attachments,
    }

    return json.dumps(serialized)


def export_soi(soi, compressed=True):
    """Export SOI.

    A .txt-file is created to contain compressed SOI.
    A .json-file is created to contain uncompressed SOI.

    The generated file-name will be on the format: "SOI_title_YYYY_mm_dd",
    where title is the SOI-title.

    Parameters
    ----------
    soi: soitool.soi.SOI
        SOI to export
    compressed : bool, optional
        Serialized SOI will be compressed if True (default).
    """
    # Serialize SOI
    serialized = serialize_soi(soi)

    # Generate filename
    title = soi.title
    date = datetime.now().strftime("%Y_%m_%d")
    file_name = f"SOI_{title}_{date}"

    if compressed:
        serialized = compress(serialized)
        file = open(file_name + ".txt", "w")
    else:
        file = open(file_name + ".json", "w")

    file.write(str(serialized))
    file.close()


def import_soi(file_path):
    """Import compressed or uncompressed serialized SOI.

    Reads content of file and decompresses it for .txt-files.
    Creates an SOI-object based on the file content.

    Parameters
    ----------
    file_path : string
        Full path to a file containing serialized SOI.

    Returns
    -------
    soitool.soi.SOI
        SOI-object

    Raises
    ------
    ValueError
        If file content is invalid against SERIALIZED_SOI_SCHEMA.
    TypeError
        If 'type' of module or attachment is not implemented.
    """
    with open(file_path, "r") as file:
        if file_path.endswith(".txt"):
            serialized = json.loads(decompress(file.read()))
        else:
            serialized = json.load(file)

    # Raise error if file content is invalid
    if not SERIALIZED_SOI_SCHEMA.is_valid(serialized):
        raise ValueError("Serialized SOI does not have correct format.")

    # Create dict for modules with instantiated widget(s)
    modules = []
    for module in serialized["modules"]:
        module_type = module["type"]

        if module_type == "TableModule":
            size = module["size"]
            data = module["data"]
            modules.append(
                {"widget": TableModule(size, data), "meta": module["meta"]}
            )
        elif module_type == "AuthenticationBoardModule":
            size = module["size"]
            data = module["data"]
            modules.append(
                {
                    "widget": AuthenticationBoardModule(size, data),
                    "meta": module["meta"],
                }
            )
        elif module_type == "SubtractorcodesModule":
            size = module["size"]
            data = module["data"]
            modules.append(
                {
                    "widget": SubtractorcodesModule(size, data),
                    "meta": module["meta"],
                }
            )
        elif module_type == "FreeTextModule":
            size = module["size"]
            data = module["data"]
            modules.append(
                {"widget": FreeTextModule(size, data), "meta": module["meta"]}
            )
        else:
            raise TypeError(
                "Module-type '{}' is not recognized.".format(module_type)
            )

    # Create dict for attachments with instantiated widget(s)
    attachments = []
    for attachment in serialized["attachments"]:
        module_type = attachment["type"]

        if module_type == "TableModule":
            size = attachment["size"]
            data = attachment["data"]
            attachments.append(
                {
                    "widget": TableModule(size, data),
                    "meta": attachment["meta"],
                }
            )
        else:
            raise TypeError(
                "Module-type '{}' is not recognized.".format(module_type)
            )
    # Create SOI
    soi = SOI(
        serialized["title"],
        serialized["description"],
        serialized["version"],
        serialized["date"],
        serialized["valid"]["from_date"],
        serialized["valid"]["to_date"],
        serialized["icon"],
        serialized["classification"],
        serialized["orientation"],
        serialized["placement_strategy"],
        serialized["algorithm_bin"],
        serialized["algorithm_pack"],
        serialized["algorithm_sort"],
        modules,
        attachments,
    )

    return soi
